/* fast_bokeh_slickgrid.css */
.bk-root .slick-header-column.ui-state-default {
  border-right: 1px solid var(--neutral-focus);
}

.bk-root .slick-sort-indicator-numbered {
  color: #6190CD;
}

.bk-root .slick-sortable-placeholder {
  background: var(--neutral-focus);
}

.bk-root .slick-cell,
.bk-root .slick-headerrow-column,
.bk-root .slick-footerrow-column {
  border: 1px solid transparent;
  border-right: 1px dotted var(--neutral-focus);
  border-bottom-color: var(--neutral-focus);
}

.bk-root .slick-cell,
.bk-root .slick-headerrow-column {
  border-bottom: none;
}

.bk-root .slick-footerrow-column {
  border-top-color: var(--neutral-focus);
}

.bk-root .slick-cell.highlighted {
  background: lightskyblue;
  background: rgba(0, 0, 255, 0.2);
}

.bk-root .slick-cell.flashing {
  border: 1px solid red !important;
}

.bk-root .slick-cell.editable {
  background: white;
  border-color: black;
  border-style: solid;
}
.bk-root .slick-cell.bk-cell-index {
  background: transparent;
}
.bk-root .slick-reorder-proxy {
  background: blue;
  opacity: 0.15;
}

.bk-root .slick-reorder-guide {
  background: blue;
  opacity: 0.7;
}

.bk-root .slick-selection {
  border: 2px dashed black;
}

.bk-root .slick-header-columns {
  border-bottom: 1px solid var(--accent-fill-rest);
}

.bk-root .slick-header-column {
  border-right: 1px solid var(--neutral-focus);
  background: transparent;
  font-weight: 500px;
}

.bk-root .slick-header-column:hover,
.bk-root .slick-header-column-active {
  background: transparent url('images/header-columns-over-bg.gif') repeat-x center bottom;
}

.bk-root .slick-headerrow {
  background: #fafafa;
}

.bk-root .slick-headerrow-column {
  background: #fafafa;
}

.bk-root .slick-row.ui-state-active {
  background: #F5F7D7;
}

.bk-root .slick-row {
  background: transparent;
}

.bk-root .slick-row.selected {
  background: var(--accent-fill-active);
}

.bk-root .slick-row:hover, .bk-root .slick-row.odd:hover {
  background-color: var(--accent-fill-hover);
  color: var(--accent-foreground-cut-rest);
}

.bk-root .slick-group {
  border-bottom: 2px solid var(--neutral-focus);
}

.bk-root .slick-group-totals {
  color: gray;
  background: white;
}

.bk-root .slick-cell.selected {
  background-color: var(--accent-fill-rest);
  color: var(--accent-foreground-cut-rest);
}

.bk-root .slick-cell.active {
  background-color: var(--accent-fill-active);
  border: none;
  border-bottom: 1px solid var(--accent-fill-active);
}

.bk-root .slick-sortable-placeholder {
  background: var(--neutral-focus) !important;
}

.bk-root .slick-row.odd {
  background: var(--neutral-fill-active);
}

.bk-root .slick-row.ui-state-active {
  background: #F5F7D7;
}

.bk-root .slick-row.loading {
  opacity: 0.5;
}

.bk-root .slick-cell.invalid {
    border-color: red;
}

@-moz-keyframes slickgrid-invalid-hilite {
  from {
    box-shadow: 0 0 6px red;
  }
  to {
    box-shadow: none;
  }
}

@-webkit-keyframes slickgrid-invalid-hilite {
  from {
    box-shadow: 0 0 6px red;
  }
  to {
    box-shadow: none;
  }
}

.bk-root .slick-header-menuitem-disabled {
    color: var(--neutral-focus);
}

.bk-root .slick-columnpicker {
  background: #f0f0f0;
}

.bk-root .slick-columnpicker li {
  background: none;
}

.bk-root .slick-columnpicker li a:hover {
  background: white;
}

.bk-root .slick-pager .ui-icon-container {
  border-color: gray;
}

.bk-root .bk-cell-special-defaults {
  border-right-color: var(--neutral-focus);
  border-right-style: solid;
  background: #f5f5f5;
}

.bk-root .bk-cell-select {
  border-right-color: var(--neutral-focus);
  border-right-style: solid;
  background: #f5f5f5;
}

.bk-root .bk-cell-index {
  border-right-color: var(--neutral-focus);
  border-right-style: solid;
  background: #f5f5f5;
  color: gray;
}
