{% extends 'base.html' %}
{% load static %}

{% block content %}

    <div class="jumbotron jumbotron-fluid bg-success text-center">
        <h1>Sliders App</h1>
        <p class="lead">Panel Sinewave</p>
    </div>

    <div class="container-fluid text-center">
        <div class="row">
            <div class="col-md-6">
                <h1>This is App1 </h1>
                {{ script | safe }}
            </div>
            <div class="col-md-6">
                <h1>This is App2 </h1>
                {{ script | safe }}
            </div>
        </div>
    </div>

{%  endblock content %}
