@import url("https://golden-layout.com/assets/css/goldenlayout-dark-theme.css");

.lm_content {
  border: 1px solid black;
  border-top-color: black;
  border-top-style: solid;
  border-top-width: 1px;
  border-right-color: black;
  border-right-style: solid;
  border-right-width: 1px;
  border-bottom-color: black;
  border-bottom-style: solid;
  border-bottom-width: 1px;
  border-left-color: black;
  border-left-style: solid;
  border-left-width: 1px;
  border-image-source: initial;
  border-image-slice: initial;
  border-image-width: initial;
  border-image-outset: initial;
  border-image-repeat: initial;
}

.lm_header .lm_tab.lm_active {
  padding-bottom: 6px;
}

